#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from itertools import chain
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Sequence
from typing import Tuple
import sys

from .parsing import BindParams, compile_bind_parameters
from . import exceptions


known_styles: Dict[type, str] = {}


def get_param_style(conn: Any) -> str:

    conncls = conn.__class__
    try:
        return known_styles[conncls]
    except KeyError:
        modname = conncls.__module__
        while modname:
            try:
                style = sys.modules[modname].paramstyle  # type: ignore
                known_styles[conncls] = style
                return style
            except AttributeError:
                if "." in modname:
                    modname = modname.rsplit(".", 1)[0]
                else:
                    break
    raise TypeError(f"Can't find paramstyle for connection {conn!r}")


class Query:

    name: str
    metadata: Mapping
    sql: str
    source: str

    def __init__(self, name, metadata, statements, source):
        self.name = name
        self.metadata = metadata
        self.statements = statements
        self.source = source
        self._conn = None

    def prepare(self, paramstyle, kw: Mapping) -> List[Tuple[str, BindParams]]:
        return [
            compile_bind_parameters(paramstyle, s, kw) for s in self.statements
        ]

    def bind(self, conn) -> "Query":
        """
        Return a copy of the query bound to a database connection
        """
        cls = self.__class__
        bound = cls.__new__(cls)
        bound.__dict__ = self.__dict__.copy()
        bound._conn = conn
        return bound

    def one(self, conn=None, *, debug=False, **kwargs):
        return self(conn, debug=debug, _result="one", **kwargs)

    def first(self, conn=None, *, debug=False, **kwargs):
        return self(conn, debug=debug, _result="first", **kwargs)

    def one_or_none(self, conn=None, *, debug=False, **kwargs):
        return self(conn, debug=debug, _result="one_or_none", **kwargs)

    def many(self, conn=None, *, debug=False, **kwargs):
        return self(conn, debug=debug, _result="many", **kwargs)

    def scalar(self, conn=None, *, debug=False, **kwargs):
        return self(conn, debug=debug, _result="scalar", **kwargs)

    def affected(self, conn=None, *, debug=False, **kwargs):
        return self(conn, debug=debug, _result="affected", **kwargs)

    def column(self, conn=None, *, debug=False, **kwargs):
        return self(conn, debug=debug, _result="column", **kwargs)

    def cursor(self, conn=None, *, debug=False, **kwargs):
        return self(conn, debug=debug, _result="cursor", **kwargs)

    def __call__(self, conn=None, *, debug=False, _result=None, **kw):
        if conn is None:
            conn = self._conn
            if conn is None:
                raise TypeError(
                    "Query must be called with a connection argument"
                )
        rt = _result or self.metadata["result"]

        paramstyle = get_param_style(conn)
        cursor = conn.cursor()

        for sqltext, bind_params in self.prepare(paramstyle, kw):
            if debug:
                import textwrap

                print(
                    f"Executing \n{textwrap.indent(sqltext, '    ')} with {bind_params!r}",
                    file=sys.stderr,
                )
            try:
                cursor.execute(sqltext, bind_params)
            except BaseException:
                _handle_exception(conn)

        if rt == "one":
            row = cursor.fetchone()
            if row is None:
                raise exceptions.NoResultFound()
            if cursor.fetchone() is not None:
                raise exceptions.MultipleResultsFound()
            return row

        if rt == "first":
            return cursor.fetchone()

        if rt == "many":
            return iter(cursor.fetchone, None)

        if rt == "one_or_none":
            row = cursor.fetchone()
            if cursor.fetchone() is not None:
                raise exceptions.MultipleResultsFound()
            return row

        if rt == "scalar":
            result = cursor.fetchone()
            if result is None:
                raise exceptions.NoResultFound()
            if isinstance(result, Mapping):
                return next(iter(result.values()))
            elif isinstance(result, Sequence):
                return result[0]
            raise TypeError(
                f"Can't find first column for row of type {type(row)}"
            )

        if rt == "column":
            first = cursor.fetchone()
            if first:
                if isinstance(first, Mapping):
                    key = next(iter(first))
                elif isinstance(first, Sequence):
                    key = 0
                else:
                    raise TypeError(
                        f"Can't find first column for row of type {type(row)}"
                    )

                return (
                    row[key]
                    for row in chain([first], iter(cursor.fetchone, None))
                )
            return iter([])

        if rt == "affected":
            return cursor.rowcount

        if rt == "cursor":
            return cursor

        raise ValueError(f"Unsupported result type: {rt}")


def _handle_exception(conn):
    """
    We have an exception of unknown type, probably raised
    from the dbapi module
    """
    exc_type, exc_value, exc_tb = sys.exc_info()
    if exc_type and exc_value:
        classes = [exc_type]
        while classes:
            cls = classes.pop()
            clsname = cls.__name__

            if clsname in exceptions.pep_249_exception_names:
                newexc = exceptions.pep_249_exception_names[clsname]()
                newexc.args = getattr(exc_value, "args", tuple())
                raise newexc.with_traceback(exc_tb) from exc_value
            classes.extend(getattr(cls, "__bases__", []))

        raise exc_value.with_traceback(exc_tb)



#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pathlib
from itertools import chain
from itertools import count
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Union

import wrapt

from .query import Query
from .exceptions import InvalidStatement
from .parsing import DEFAULT_RESULT_TYPE
from .parsing import split_statements

SQL_FILE_GLOB = "**/*.sql"


class Module:

    reloader: Optional[Callable] = None
    _conn = None
    queries: Dict = {}

    def __init__(self):
        self.queries = {}

    def __getattr__(self, name):
        if self.reloader:
            self.reloader()
            try:
                return reloadable_query_proxy(self, name, self.reloader)
            except KeyError:
                raise AttributeError(name)
        try:
            return self.queries[name]
        except KeyError:
            raise AttributeError(name)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return

    def set_reloader(self, fn):
        self.reloader = fn

    def clear(self):
        self.queries.clear()

    def add_query(self, name, query):
        if name in self.queries:
            raise InvalidStatement(
                f"Can't add query {name!r} in {query.source}: "
                f"{self!r} already has an attribute named {name!r} "
                f"(loaded from {getattr(self, name).source})"
            )
        if self._conn:
            query = query.bind(self._conn)
        self.queries[name] = query

    def add_queries(
        self, qs: Union[Iterable[Tuple[str, Query]], Mapping[str, Query]]
    ):
        if isinstance(qs, Mapping):
            qs = qs.items()
        for name, query in qs:
            self.add_query(name, query)

    def load_dir(self, path: Union[str, pathlib.Path]):
        for p in pathlib.Path(path).glob(SQL_FILE_GLOB):
            self.load_file(p)

        return self

    def load_file(self, path: pathlib.Path):
        queries = list(load_queries(path))
        for query in queries:
            self.add_query(query.name, query)

    def execute(self, conn, sql=None, result="many", **kw):
        if sql is None:
            conn, sql = self._conn, conn
        query = Query(
            name=None,
            metadata={"result": result},
            statements=[sql],
            source="<string>",
        )
        return query(conn, **kw)

    def one(self, conn, sql=None, **kwargs):
        return self.execute(conn, sql, result="one", **kwargs)

    def one_or_none(self, conn, sql=None, **kwargs):
        return self.execute(conn, sql, result="one_or_none", **kwargs)

    def many(self, conn, sql=None, **kwargs):
        return self.execute(conn, sql, result="many", **kwargs)

    def scalar(self, conn, sql=None, **kwargs):
        return self.execute(conn, sql, result="scalar", **kwargs)

    def affected(self, conn, sql=None, **kwargs):
        return self.execute(conn, sql, result="affected", **kwargs)

    def column(self, conn, sql=None, **kwargs):
        return self.execute(conn, sql, result="column", **kwargs)

    def cursor(self, conn, sql=None, **kwargs):
        return self.execute(conn, sql, result="cursor", **kwargs)

    def bind(self, conn) -> "Module":
        """
        Return a copy of the module bound to a database connection
        """
        cls = self.__class__
        bound = cls.__new__(cls)
        bound.__dict__ = {
            "_conn": conn,
            "reloader": self.reloader,
            "queries": {
                name: q.bind(conn) for name, q in self.queries.items()
            },
        }
        return bound

    def transaction(self, conn) -> "Transaction":
        return Transaction(self, conn)

    def savepoint(self, conn) -> "Savepoint":
        return Savepoint(self, conn)


class Transaction:
    def __init__(self, module, conn):
        self.conn = conn
        self.module = module.bind(conn)

    def __enter__(self):
        return self.module

    def __exit__(self, type, value, traceback):
        if type:
            self.conn.rollback()
        else:
            self.conn.commit()

    def commit(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()


class Savepoint:

    _seq = count()

    def __init__(self, module, conn):
        self.conn = conn
        self.module = module.bind(conn)
        self.savepoint = f"sp_{next(self._seq)}"
        self._cursor = None

    def __enter__(self):
        self._cursor = self.conn.cursor()
        self._cursor.execute(f"SAVEPOINT {self.savepoint}")
        return self.module

    def __exit__(self, type, value, traceback):
        if self._cursor is not None:
            if type:
                self._cursor.execute(f"ROLLBACK TO SAVEPOINT {self.savepoint}")
            else:
                self._cursor.execute(f"RELEASE SAVEPOINT {self.savepoint}")
            self._cursor.close()


def module(path, auto_reload=False):
    path = pathlib.Path(path)
    module = Module()
    module.load_dir(path)
    if auto_reload:

        def get_all_mtimes(path):
            return chain(
                [path.stat().st_mtime],
                (p.stat().st_mtime for p in path.glob(SQL_FILE_GLOB)),
            )

        module_mtime = [max(get_all_mtimes(path))]

        def reload_module():
            if any(m > module_mtime[0] for m in get_all_mtimes(path)):
                module_mtime[:] = [max(get_all_mtimes(path))]
                module.clear()
                module.load_dir(path)
                return True
            return False

        module.set_reloader(reload_module)
    return module


def load_queries(path: pathlib.Path) -> Iterable[Query]:
    with path.open("r", encoding="UTF-8") as f:
        sql = f.read()
        for ix, (metadata, statements) in enumerate(split_statements(sql)):
            metadata.setdefault("result", DEFAULT_RESULT_TYPE)
            if not metadata.get("name"):
                if ix == 0:
                    metadata["name"] = path.stem
                else:
                    raise InvalidStatement(
                        f"{path!s}: no name specified (eg `-- :name my_query_name`)"
                    )
            if "result" not in metadata:
                raise InvalidStatement(
                    f"{path!s}: no result type specified (eg `-- :result :many`)"
                )
            yield Query(
                metadata["name"], metadata, statements, source=str(path)
            )


def reloadable_query_proxy(module, name, reloader):
    def get_query():
        return module.queries[name]

    class ReloadableQueryProxy(wrapt.ObjectProxy):
        def __call__(self, *args, **kwargs):
            reloader()
            self.__wrapped__ = get_query()
            return self.__wrapped__(*args, **kwargs)

        def bind(self, conn):
            return reloadable_query_proxy(module, name, reloader)

    return ReloadableQueryProxy(get_query())



from itertools import chain
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Iterable
from typing import Sequence
from typing import Set


def fetch_dicts(cursor) -> Iterable[Dict]:
    """
    Return an iterable over db-api cursor rows. Each row is returned as a
    dict.
    """
    keys = [item[0] for item in cursor.description]
    row: Sequence
    for row in iter(cursor.fetchone, None):
        yield dict(zip(keys, row))


def group_and_aggregate(items: Iterable[Dict[str, Any]], **groups: Sequence):
    """
    Aggregate and nest joined records in query results.

    Example:

        >>> import pprint
        >>> data = [
        ...     {'id': 1, 'name': 'fred', 'email': 'fred@example.org'},
        ...     {'id': 1, 'name': 'jim', 'email': 'jim@example.org'},
        ...     {'id': 2, 'name': 'sheila', 'email': 'sheila@example.org'},
        ... ]
        >>> pprint.pprint(list(group_and_aggregate(data, people=('email', 'name'))))
        [{'id': 1,
          'people': [{'email': 'fred@example.org', 'name': 'fred'},
                     {'email': 'jim@example.org', 'name': 'jim'}]},
          {'id': 2, 'people': [{'email': 'sheila@example.org', 'name': 'sheila'}]}]
    """

    DEFAULT = ""
    items = iter(items)
    peek = next(items, None)
    if peek is None:
        return
    group_items: List[Tuple[str, Set[str]]] = [
        (k, set(v)) for k, v in groups.items()
    ]
    allgroupkeys = {colname for k, v in group_items for colname in v}
    group_items = [
        (DEFAULT, {k for k in peek if k not in allgroupkeys})
    ] + group_items
    items = chain([peek], items)

    lastrow = None
    record: Dict[Any, Any] = {}
    row: Dict[Any, Any]
    for row in items:
        current = record
        is_new = False
        for groupname, groupkeys in group_items:
            group = {k: row[k] for k in groupkeys}
            is_new = (
                is_new
                or lastrow is None
                or any(lastrow[k] != v for k, v in group.items())
            )
            if is_new:
                if groupname is DEFAULT and record:
                    yield record[DEFAULT][0]
                    current = record = {}
                current.setdefault(groupname, []).append(group)
            current = current[groupname][-1]

        lastrow = row

    if record:
        yield record[DEFAULT][0]


if __name__ == "__main__":
    data = [
        {"id": 1, "name": "fred", "email": "fred@example.org"},
        {"id": 1, "name": "jim", "email": "jim@example.org"},
        {"id": 2, "name": "sheila", "email": "sheila@example.org"},
    ]
    import pprint

    pprint.pprint(list(group_and_aggregate(data, people=("email", "name"),)))
